/**
 * Programm that implements the basic FIFO-operations and tests them (init, empty, full, enqueue, dequeue)
 * 
 * @author robert.binna@uibk.ac.at
 * @date 09.04.2008
 */
 
#include <stdio.h>
#include <assert.h>
 
#define BUFFERSIZE 30

#ifndef TRUE
#define TRUE 1
#endif

#ifndef FALSE
#define FALSE 0
#endif

struct fifo {
	int first; /* first index, the index where to insert the next element enqueue */
	int last; /* last index, the index where the next element is returned dequeue */
	int isFull; /* flag that is used to distinguish between empty and full */
	int buffer[BUFFERSIZE];
};

struct fifo global_fifo = { 0 , 0, 0 };
 
/**
 * Initializes the FIFO resets it
 */
void init(struct fifo* aFifo) {
	aFifo->first = 0;
	aFifo->last = 0;
	aFifo->isFull = 0;
}

/**
 * Checks if the FIFO is empty at the call time
 *
 * @return if it is empty
 */
int empty(struct fifo* aFifo) {
	return !aFifo->isFull && aFifo->first == aFifo->last;  
}

/**
 * Checks if the FIFO is full and nothing can get inserted anymore
 *
 * @return if it is full
 */
int full(struct fifo* aFifo) {
	return aFifo->isFull;
}

/**
 * Dequeue adds an element to the tail of the queue. It is responsible for returning an error if
 * the queue is full
 *
 * @returns 1 if all is okay 0 if the the FIFO is full
 */
int enqueue(struct fifo* aFifo, int element) {
	int success = TRUE;
	if(full(aFifo)) {
		success = FALSE;
	} else {
		aFifo->buffer[aFifo->first] = element;
		++aFifo->first;
		aFifo->first%=BUFFERSIZE;
		aFifo->isFull = (aFifo->first == aFifo->last);
	}
	return success;
}

/**
 * Dequeue removes the first element from the fifo queue. It is responsible for returning an error if
 * the queue is empty
 *
 * @param output an output parameter. It will be filled with the element that is removed from 
 * the FIFO. If the buffer is empty the value will be the same as before
 * @return 1 if removal of the first element was okay otherwise 0. 0 indicates that the FIFO was already empty.
 */
int dequeue(struct fifo* aFifo, int* output) {
	int success = TRUE;
	if(empty(aFifo)) {
		success = FALSE;
	} else {
		*output = aFifo->buffer[aFifo->last];
		++aFifo->last;
		aFifo->last%=BUFFERSIZE;
		aFifo->isFull = FALSE;
	}
	return success;
}

/**
 * main function for testing the fifo operations
 * tests the correct behaviour of all base operations. The behaviour is checked through 
 * assert macro calls.
 * @return the exit value of the program. 
 * 0 for correct behaviour.
 */
int main(void) {
	int i=0;

	init(&global_fifo);

	printf("FIFO empty (should be 1): %d\n", empty(&global_fifo));
	assert(empty(&global_fifo));
	printf("FIFO full  (should be 0): %d\n", full(&global_fifo));
	assert(!full(&global_fifo));
	
	for(i=0; i < BUFFERSIZE; ++i) {
		assert(!full(&global_fifo));
		printf("Enqueue %d\n",i);
		assert(enqueue(&global_fifo,i));
		assert(!empty(&global_fifo));
	}
	
	assert(full(&global_fifo));
	assert(!enqueue(&global_fifo,i));
	
	for(i=0; i < BUFFERSIZE; ++i) {
		assert(!empty(&global_fifo));
		int output = 0;
		printf("Dequeue %d",i);
		assert(dequeue(&global_fifo,&output));
		printf("Got Value %d\n",output);
		assert(output == i);
		assert(!full(&global_fifo));
	}
	
	printf("FIFO empty (should be 1): %d\n", empty(&global_fifo));
	assert(empty(&global_fifo));
	printf("FIFO full  (should be 0): %d\n", full(&global_fifo));
	assert(!full(&global_fifo));
	
	assert(!dequeue(&global_fifo,&i));
	
	init(&global_fifo);

	printf("FIFO empty (should be 1): %d\n", empty(&global_fifo));
	assert(empty(&global_fifo));
	printf("FIFO full  (should be 0): %d\n", full(&global_fifo));
	assert(!full(&global_fifo));
	
	for(i=0; i < BUFFERSIZE; ++i) {
		assert(!full(&global_fifo));
		printf("Enqueue %d\n",i);
		assert(enqueue(&global_fifo, i));
		assert(!empty(&global_fifo));
	}
	
	assert(full(&global_fifo));
	assert(!enqueue(&global_fifo,i));
	
	init(&global_fifo);
	
	printf("FIFO empty (should be 1): %d\n", empty(&global_fifo));
	assert(empty(&global_fifo));
	printf("FIFO full  (should be 0): %d\n", full(&global_fifo));
	assert(!full(&global_fifo));
	
	return 0;
}
